#include "evalDet.hpp"

/*
    MOT17 row format:
0,     1,  2,   3,    4,     5,      6,         7,        8
frame, id, left, top, width, height, conf-score, world_x, world_y 

 */

void evalDet(vector<string>& sequences, VVVD& pre_proc_dets) {

    VVVD dets(pre_proc_dets.size());
    VVVD rawGT(pre_proc_dets.size());

    vector<string> gt_paths = insert_seq_path("data/", sequences, "/gt/gt.txt");

    load_gt(gt_paths, rawGT);
    cleanData(rawGT, pre_proc_dets, dets, VIS_THRESH);
    
    // assert same number of frames
    assert(rawGT.size() == dets.size());

    VVVD gt;
    gt.resize(dets.size());

    cleanGT(rawGT, gt, VIS_THRESH);
    VI gt_matches;
    VVD det_matches;
    matchDet(gt, dets, gt_matches, det_matches);

    // x_coord reference points for detection P/R curve
    VD ref = {0, .25, .50, .75, 1};
    VD recall_axis;
    VD precision_axis;
    compPR(gt_matches, det_matches, ref, recall_axis, precision_axis);

    /*
     * Do something with recall/precision curve?
     */
    double AP = sum(ref)/(double)ref.size();;

    cout << "Detection quality results (higher is better, 100 max): " << endl << endl;
    //cout << "\taverage precision: " << AP << endl;

    clear_mod_hun(gt, dets);
}

// Helper function that computes area overlap of bounding boxes 
double compOa(VD& gt, VD& det, bool ignore) {

    double w = std::min(det[4] + det[2], gt[4] + gt[2]) - std::max(det[2],gt[2]);
    if(w <= 0) 
        return 0.0;

    double h = std::min(det[5] + det[3], gt[5] + gt[3]) - std::max(det[3],gt[3]); 
    if(h <= 0)
        return 0.0;

    double i=w*h;
    double u; 

    if(ignore)
        u = det[4]*det[5]; 
    else 
        u = det[4]*det[5]+gt[4]*gt[5]-i; 

    return i/u;
}

// gt/det are mats of detetcion for one frame
// gt_match is corresponding det mathc, and det_match is gt match + score
// previously resEval in MOT-devkit
void matchDet(VVVD& gt, VVVD& det, VI& gt_match, VVD& det_match) {
    
    byScore comp(6);
    for(size_t frame = 0; frame < gt.size(); ++frame) {
        
        VVD& curGT = gt[frame];
        VVD& curDet = det[frame];  
        std::sort(curDet.begin(), curDet.end(), comp);

        // if ith det or jth gt has a match, a 1 wil be at f_det[i] / f_gt[j], otherwise will be 0 for unmatched, or -1 for taken match
            // Note f_det has an extra field to keep old score [score x match_flag] (used for sorting in compRP)
        VI f_gt;
        f_gt.resize(curGT.size(), 0);
        
        VVD f_det(curDet.size(), {0,0});

        for(size_t d = 0; d < curDet.size(); ++d) {
            double bestOa = 0.5; //threshold
            int bestGT, bestFlag = 0;
            // match flag; -1 for ignore gt or unmatchable det, 0 for currently unmatched, 1 for matched
            int flag;

            for(size_t g = 0; g < curGT.size(); ++g) {
                // skip this gt if already matched
                assert(curGT[g][6] != -1);
                flag = f_gt[g];
                if(flag == 1)
                    continue;
                // break if det is matched and at ignored gt, won't find better gt
                if(bestFlag != 0 && flag == -1)
                    break;  
                double oa = compOa(curGT[g], curDet[d], false);
                if(oa < bestOa)
                    continue;

                bestOa = oa;
                bestGT = g;
                bestFlag = ( (flag == 0) ? 1 : -1);
            }
            //save score
            f_det[d][0] = curDet[d][6];
            f_det[d][1] = bestFlag;
            if(bestFlag == 1) 
                f_gt[bestGT] = bestFlag; 
        }
        for(int match : f_gt)
            gt_match.emplace_back(match);

        for(VD& match : f_det)
            det_match.emplace_back(std::move(match));
    }
}

// gt and dt should be complete (all seqs) outputs of matchDet. 
// ref is desired x coords for PR curve, default [0, .25, .50, .75, 1]
void compPR(VI& gt, VVD& dt, VD& ref, VD& xs, VD& ys) {
// Compute PR based on outputs of matchDet on multiple images.
//
// Also computes result at reference points (ref):
//  which for PR curves is the *precision* at reference *recall*
//
// INPUTS
//  gt         - first output of matchDet() for each image: holds index to best det or -1/0 for no match
//  dt         - second output of matchDet() for each image: holds [confidence score, best index to gt]
//  ref        - [] reference points for PR curve
//
// OUTPUTS
//  xs         - x coords for curve: PR->recall
//  ys         - y coords for curve: PR->precision
//  score      - detection scores corresponding to each (x,y)
//  ref        - recall or precision at each reference point

    if(dt.empty()) { 
        return; 
    }
    std::sort(dt.begin(), dt.end(), byScore(0));

    size_t np = gt.size(); 

    VD score;
    VI tp;
    for(auto & row : dt) {
        // NOTE: 0 and 1 are match indeces.
        tp.emplace_back(row[1]);
    }
    
    //fp=double(tp~=1); 
    VI fp;
    for(auto& val : tp){
        assert(val != -1);
        int false_positive = ((val != 1) ? 1 : 0);
        fp.emplace_back(false_positive);
    }
    //fp=cumsum(fp);
    int tempTotal = 0;
    for(size_t i = 0; i < fp.size(); ++i) {
        tempTotal += fp[i];
        fp[i] = tempTotal;
    } 

    //tp=cumsum(tp);
    tempTotal = 0;
    for(size_t i = 0; i < tp.size(); ++i) {
        tempTotal += tp[i];
        tp[i] = tempTotal;
    } 

    // xs=tp/np; // holds recall
    for(int val : tp)
        xs.emplace_back( ((double)val)/((double)np));
    
    //ys=tp./(fp+tp); // holds precision
    for(size_t i = 0; i < tp.size(); ++i)
        ys.emplace_back(((double)tp[i])/((double)tp[i]+fp[i]));

    //xs1=[xs; inf]; 
    VD xs1(xs);
    xs1.emplace_back(DBL_MAX);
    
    //ys1=[ys; 0];
    VD ys1(ys);
    ys1.emplace_back(0);

    //for i=1:m,
    for(size_t i = 0; i < ref.size(); ++i) {     
        //j = find(xs1>=ref(i)); ref(i)=ys1(j(1));
        for(size_t j = 0; j < xs1.size(); ++j) {
            if(xs1[j] >= ref[i]) {
                ref[i] = ys1[j];
                break;
            }
        }
    }
}

void clear_mod_hun(VVVD& gt, VVVD& det){

    assert(gt.size() == det.size());
    VI fp;
    VI misses;
    VI numMatches;
    int numGT = 0;
    vector<vector<double>> ious(gt.size());
    for (size_t frame = 0; frame < gt.size(); ++frame){

        if(det[frame].empty() || gt[frame].empty())
            continue;
    
        VI assignment;
        boxHungarian(gt[frame], det[frame], assignment);
        size_t nMatch = 0;
        
        for (size_t d = 0; d < det[frame].size(); ++d) {

            int g = assignment[d];
            if(g != -1) {
                ++nMatch;
                // recalc boxiou for MODP
                double val = 1 - calcBoxWeights(gt[frame][g], det[frame][d]);
                if (val <= VIS_THRESH)
                    ious[frame].emplace_back(1 - val);
            }
        }
        numMatches.emplace_back(nMatch);
        fp.emplace_back(det[frame].size() - nMatch);
        misses.emplace_back(gt[frame].size() - nMatch);
        numGT += gt[frame].size();
    }

    int missed = sum(misses);
    int falsePositives = sum(fp);
    int truePositives = sum(numMatches);

    double MODP = sum(ious)/(double)truePositives * 100;
    double MODA = (1.0f-((double)missed+falsePositives)/(numGT))*100;
    double recall = ((double)truePositives)/numGT*100;
    double precision = truePositives/((double)(falsePositives) + truePositives) * 100;
    double FAR = (double)(falsePositives)/gt.size();

    cout << "\tRecall (correct detections / ground-truth detections): " << recall << endl;
    cout << "\tPrecision (true-positive detections / total detections): " << precision << endl;
    /*cout << "\tFAR: " << FAR << endl;
    cout << "\tNum GT (very sensitive to VIS_THRESH): " << numGT << endl;
    cout << "\ttrue Positives: " << truePositives << endl;
    cout << "\tfalse Positives: " << falsePositives << endl;
    cout << "\tfalse negatives: " << missed << endl;*/
    cout << "\tDetection accuracy (incl false positives, missed targets): " << MODA << endl;
    cout << "\tDetection precision (bounding box alignment vs ground-truth): " << MODP << endl << endl;
}
