#include "kalman_box_tracker.hpp"
using cv::KalmanFilter;
using cv::Mat;
using cv::Mat_;
using std::vector;

using namespace cv;
using namespace std;

/*
Kalman notes:
kf.x: initial positions and velocities
kf.F: state transition matrix (dim_x by dim_x)
kf.H: measurement function (dim_x by dim_z)
kf.P: covariance mtx. (dim_x by dim_x)
kf.Q: process noise.
kf.R: measurement noise (dim z by dim z)
*/

Mat convert_bbox_to_z(const vector<float> &bbox, bool update){
    //Takes a bounding box in the form [x1,y1,x2,y2] and returns z in the form
    //[x,y,s,r] where x,y is the centre of the box and s is the scale/area and r is
    //the aspect ratio
    float w = bbox[2] - bbox[0];
    float h = bbox[3] - bbox[1];
    float x = bbox[0] + (w / 2);
    float y = bbox[1] + (h / 2);
    float s = w*h; //scale is area
    float r = w / h;
    
    if (update) {
        return (Mat_<float>(4,1) << x, y, s, r);
    }
    else {
        return (Mat_<float>(7,1) << x, y, s, r, 0, 0, 0);
    }
}

vector<float> convert_x_to_bbox(Mat x) {
    float x0 = x.at<float>(0,0), x1 = x.at<float>(1,0), x2 = x.at<float>(2,0), x3 = x.at<float>(3,0);
    float w = sqrt(x2 * x3);
    float h = x2 / w;
    vector<float> results;
    results.push_back(x0 - (w/2));
    results.push_back(x1 - (h/2));
    results.push_back(x0 + (w/2));
    results.push_back(x1 + (h/2));
    return results;
}

int KalmanBoxTracker::count = 0;
KalmanBoxTracker::KalmanBoxTracker(vector<float> bbox) {
    //make the 7 x 4 KalmanFilter
    kf = KalmanFilter(7,4);
    //make the transition matrix (python: kf.F)
    kf.transitionMatrix = (Mat_<float>(7,7) <<
    1,0,0,0,1,0,0,
    0,1,0,0,0,1,0,
    0,0,1,0,0,0,1,
    0,0,0,1,0,0,0,
    0,0,0,0,1,0,0,
    0,0,0,0,0,1,0,
    0,0,0,0,0,0,1
    );
    //measurement matrix (python kf.H)
    kf.measurementMatrix = (Mat_<float>(4,7) <<
    1,0,0,0,0,0,0,
    0,1,0,0,0,0,0,
    0,0,1,0,0,0,0,
    0,0,0,1,0,0,0
    );
    //measurement noise (python kf.R)
    kf.measurementNoiseCov = (Mat_<float>(4,4) <<
    1,0,0,0,
    0,1,0,0,
    0,0,10,0,
    0,0,0,10
    );
    //covariance matrix (python kf.P)
    
    kf.errorCovPost = (Mat_<float>(7,7) <<
    10,0,0,0,0,0,0,
    0,10,0,0,0,0,0,
    0,0,10,0,0,0,0,
    0,0,0,10,0,0,0,
    0,0,0,0,10000,0,0,
    0,0,0,0,0,10000,0,
    0,0,0,0,0,0,10000
    );
    //process noise matrix (python kf.Q)
    kf.processNoiseCov = (Mat_<float>(7,7) <<
    1,0,0,0,0,0,0,
    0,1,0,0,0,0,0,
    0,0,1,0,0,0,0,
    0,0,0,1,0,0,0,
    0,0,0,0,0.01,0,0,
    0,0,0,0,0,0.01,0,
    0,0,0,0,0,0,0.0001
    );
    //initial position vector (python kf.x)
    kf.statePost = convert_bbox_to_z(bbox, false);

    //initialize the other things
    time_since_update = 0;
    id = KalmanBoxTracker::count;
    KalmanBoxTracker::count++;
    //history is initialized automatically to empty vector
    hits = 0;
    hit_streak = 0;
    age = 0;
}

void KalmanBoxTracker::update(vector<float> bbox){
    time_since_update = 0;
    history = vector<vector<float> >();
    hits++;
    hit_streak++;
    kf.correct(convert_bbox_to_z(bbox, true));
}

vector<float> KalmanBoxTracker::predict() {
    if (kf.statePost.at<float>(6,0) + kf.statePost.at<float>(2,0) <= 0) {
        kf.statePost.at<float>(6,0) = 0;
    }
    kf.predict();
    age += 1;
    if (time_since_update > 0) {
        hit_streak = 0;
    }
    time_since_update++;
    history.push_back(convert_x_to_bbox(kf.statePost));
    return history.back();

}

vector<float> KalmanBoxTracker::get_state() {
    return convert_x_to_bbox(kf.statePost);
}

int KalmanBoxTracker::get_hit_streak() {
    return hit_streak;
}

int KalmanBoxTracker::get_hits() {
    return hits;
}
int KalmanBoxTracker::get_time_since_update() {
    return time_since_update;
}

int KalmanBoxTracker::get_id() {
    return id;
}
