#include "linear_assignment.hpp"
#include <utility>
#include <vector>
#include <algorithm>
#include <stdexcept>
using std::vector;
using std::pair;
using std::make_pair;
using std::runtime_error;
using std::min_element;
using std::sort;

struct _HungarianState {
    bool transposed;
    vector<vector<float> > C;
    vector<bool> row_uncovered;
    vector<bool> col_uncovered;
    int Z0_r;
    int Z0_c;
    vector<vector<int> > path;
    vector<vector<int> > marked;
    _HungarianState(vector<vector<float> > cost_matrix) {
        if (cost_matrix.size() < 1 || cost_matrix[0].size() < 1) {
            throw runtime_error("Not 2 dimensional input");
        }
        transposed = (cost_matrix.size() > cost_matrix[0].size());
        if (transposed) {
            //more rows than cols. transpose
            C = vector<vector<float> >(cost_matrix[0].size(), vector<float>(cost_matrix.size(), 0));
            for (int i = 0; i < (int)cost_matrix.size(); ++i) {
                for (int j = 0; j < (int)cost_matrix[0].size(); ++j) {
                    C[j][i] = cost_matrix[i][j];
                }
            }
        }
        else {
            C = cost_matrix;
        }
        //n is rows
        int n = (int)C.size();
        //m is cols
        int m = (int)C[0].size();
        row_uncovered = vector<bool>(n, true);
        col_uncovered = vector<bool>(m, true);
        Z0_r = 0;
        Z0_c = 0;
        path = vector<vector<int> >(n + m, vector<int>(2, 0));
        marked = vector<vector<int> >(n, vector<int>(m, 0));
    }
    void _clear_covers(){
        for (int i = 0; i < (int)row_uncovered.size(); ++i){
            row_uncovered[i] = true;
        }
        for (int j = 0; j < (int)col_uncovered.size(); ++j) {
            col_uncovered[j] = true;
        }
    }
};

enum Step {
    STEP_1, STEP_3, STEP_4, STEP_5, STEP_6, DONE
};

static vector<pair<int,int> > _hungarian( vector<vector<float> > &X);
static Step next_step(Step step, _HungarianState& state);
static Step _step1(_HungarianState& state);
static Step _step3(_HungarianState& state);
static Step _step4(_HungarianState& state);
static Step _step5(_HungarianState& state);
static Step _step6(_HungarianState& state);

vector<pair<int,int> > _hungarian( vector<vector<float> > &X){
    _HungarianState state(X);
    //line 122: our program runtime errors on a 0 dimension
    Step step = STEP_1;
    while (step != DONE) {
        step = next_step(step, state);
    }
    //now all the coordinates in state.marked that are 1 represent our coordinates
    vector<pair<int,int> > results;
    for (int i = 0; i < (int)state.marked.size(); ++i) {
        for (int j = 0; j < (int)state.marked[0].size(); ++j) {
            if (state.marked[i][j] == 1) {
                //if transposed, flip the coordinates back when returning.
                results.push_back(state.transposed ? make_pair(j,i) : make_pair(i,j));
            }
        }
    }
    return results;
}

Step next_step(Step step, _HungarianState& state){
    switch(step){
        case STEP_1:
            return _step1(state);
        case STEP_3:
            return _step3(state);
        case STEP_4:
            return _step4(state);
        case STEP_5:
            return _step5(state);
        case STEP_6:
            return _step6(state);
        default:
            throw runtime_error("State Machine Error");
    };
}

Step _step1(_HungarianState& state){
    //step1: subtract minimum element from each row
    for (int i = 0; i < (int)state.C.size(); ++i) {
       float min = *min_element(state.C[i].begin(), state.C[i].end());
       for (int j = 0; j  < (int)state.C[i].size(); ++j) {
           state.C[i][j] -= min;
       }
    }
    //step 2: find a zero in resulting matrix. 
    //if there is no starred 0 in that row and col, star it
    //repeat for each 0
    for (int i = 0; i < (int)state.C.size(); ++i) {
        for (int j = 0; j < (int)state.C[i].size(); ++j) {
            
            if (state.C[i][j] == 0) {
                if (state.row_uncovered[i] && state.col_uncovered[j]){
                    state.marked[i][j] = 1;
                }
                state.row_uncovered[i] = false;
                state.col_uncovered[j] = false;
            }
        }
    }
    state._clear_covers();
    return STEP_3;
}

Step _step3(_HungarianState& state){
    //Cover each column containing a zero. If n columns are covered, the starred 
    //0s describe a complete solution. In this case, go to DONE; otherwise, step 4
    int total = 0;
    for (int j = 0; j < (int)state.C[0].size(); ++j) {
        bool starred_col = false;
        for (int i = 0; i < (int)state.C.size(); ++i) {
            if (state.marked[i][j] == 1) {
                starred_col = true; break;
            }
        }
        if (starred_col) {
            state.col_uncovered[j] = false;
            total++;
        }
    }
    return (total < (int)state.C.size() ? STEP_4 : DONE);
}

Step _step4(_HungarianState& state){
    /*Find a noncovered zero and prime it. If there is no starred zero
    in the row containing this primed zero, Go to Step 5. Otherwise,
    cover this row and uncover the column containing the starred
    zero. Continue in this manner until there are no uncovered zeros
    left. Save the smallest uncovered value and Go to Step 6.*/
    bool uncovered_zero_found = false;
    int row = -1, col = -1;
    while (true) {
        uncovered_zero_found = false;
        for (int i = 0; i < (int)state.C.size(); ++i) {
            for (int j = 0; j < (int)state.C[i].size(); ++j) {
                if (state.C[i][j] == 0 && state.row_uncovered[i] && state.col_uncovered[j]) {
                    uncovered_zero_found = true;
                    row = i; col = j;
                    break;
                }
            }
            if (uncovered_zero_found) {
                break;
            }
        }
        if (!uncovered_zero_found) {
            return STEP_6;
        }
        else {
            state.marked[row][col] = 2;
            //scan this row looking for the first marked col
            int star_col = -1;
            for (int j = 0; j < (int)state.marked[row].size(); ++j) {
                if (state.marked[row][j] == 1) {
                    star_col = j;
                    break;
                }
            }
            //row not found. go to step 5
            if (star_col == -1) {
                state.Z0_r = row;
                state.Z0_c = col;
                return STEP_5;
            }
            else {
                col = star_col;
                state.row_uncovered[row] = false;
                state.col_uncovered[col] = true;
            }
        }
    }
    //should never execute
    return DONE;
}

Step _step5(_HungarianState& state){
    /*Construct a series of alternating primed and starred zeros as follows.
    Let Z0 represent the uncovered primed zero found in Step 4.
    Let Z1 denote the starred zero in the column of Z0 (if any).
    Let Z2 denote the primed zero in the row of Z1 (there will always be one).
    Continue until the series terminates at a primed zero that has no starred
    zero in its column. Unstar each starred zero of the series, star each
    primed zero of the series, erase all primes and uncover every line in the
    matrix. Return to Step 3*/
    int count = 0;
    state.path[count][0] = state.Z0_r;
    state.path[count][1] = state.Z0_c;
    while (true) {
        int row = -1;
        for (int i = 0; i < (int)state.marked.size(); ++i) {
            if (state.marked[i][state.path[count][1]] == 1) {
                row = i;
                break;
            }
        }
        if (row == -1) {
            //could not find one
            break;
        }
        else {
            count++;
            state.path[count][0] = row;
            state.path[count][1] = state.path[count - 1][1];
        }

        //  # Find the first prime element in the row defined by the
        //  # first path step
        int col = -1;
        int pc = state.path[count][0];
        for (int j = 0; j < (int)state.marked[pc].size(); ++j) {
            if (state.marked[pc][j] == 2) {
                col = j;
                break;
            }
        }
        //should be redundant but keeping just in case
        if (state.marked[row][col] != 2) {
            col = -1;
        }
        count++;
        state.path[count][0] = state.path[count - 1][0];
        state.path[count][1] = col;

    }
    //convert paths
    for (int i = 0; i <= count; ++i) {
        if (state.marked[state.path[i][0]][state.path[i][1]] == 1) {
            state.marked[state.path[i][0]][state.path[i][1]] = 0;
        }
        else {
            state.marked[state.path[i][0]][state.path[i][1]] = 1;
        }
    }
    state._clear_covers();
    //erase all prime markings
    for (int i = 0; i < (int)state.marked.size(); ++i) {
        for (int j = 0; j < (int)state.marked[i].size(); ++j) {
            if (state.marked[i][j] == 2) {
                state.marked[i][j] = 0;
            }
        }
    }
    return STEP_3;
}

Step _step6(_HungarianState& state){
    //Add the value found in Step 4 to every element of each covered row,
    //and subtract it from every element of each uncovered column.
    //Return to Step 4 without altering any stars, primes, or covered lines.
    bool row_unc = false, col_unc = false;
    for (int i = 0; i < (int)state.row_uncovered.size(); ++i) {
        if (state.row_uncovered[i]){
            row_unc = true;
            break;
        }
    }
    for (int j = 0; j < (int)state.col_uncovered.size(); ++j) {
        if (state.col_uncovered[j]){
            col_unc = true;
            break;
        }
    }
    float* min = nullptr;
    if (row_unc && col_unc) {
        for (int i = 0; i < (int)state.C.size(); ++i) {
            for (int j = 0; j < (int)state.C[i].size(); ++j) {
               if (state.row_uncovered[i] && state.col_uncovered[j]) {
                   if (min == nullptr) {
                       min = new float(state.C[i][j]);
                   } else {
                       if (*min > state.C[i][j]) {
                           *min = state.C[i][j];
                       }
                   }
               } 
            }
        }
        for (int i = 0; i < (int)state.C.size(); ++i) {
            for (int j = 0; j < (int)state.C[i].size(); ++j) {
                //for every covered row, add min
                if (!state.row_uncovered[i]) {
                    state.C[i][j] += *min;
                }
                //for every uncovered column, subtract min
                if (state.col_uncovered[j]) {
                    state.C[i][j] -= *min;
                }
            }
        }
    }
    delete min;
    return STEP_4;
}

vector<pair<int,int> > linear_assignment( vector<vector<float> > X){
    vector<pair<int,int> > indices;
    if (X.size() > 0) {
        indices = _hungarian(X);
    }
    sort(indices.begin(), indices.end());
    return indices;
}
