#include <iostream>
#include <ctime>
#include <cstdio>

#include "dpm.hpp"
#include "sort.hpp"
#include "utils.hpp"
#include "evalDet.hpp"
#include "evalTrack.hpp"

#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
using std::vector;
using std::string;
using std::cout;
using std::endl;

int main() {
    // use sequence = {"sample"} for small tests from MOT17-02. Others (see list, below) may be downloaded and tested independently.
    vector<string> sequences = {"sample"}; // {"MOT17-02","MOT17-04","MOT17-05","MOT17-09","MOT17-10","MOT17-11","MOT17-13"}; // MOT17
    // vector<string> sequences = {"PETS09-S2L1","TUD-Campus","TUD-Stadtmitte","ETH-Bahnhof","ETH-Sunnyday","ETH-Pedcross2","KITTI-13","KITTI-17","ADL-Rundle-6","ADL-Rundle-8","Venice-2"}; // MOT15
    double track_time = 0.0;
    double detection_time = 0.0;
    double track_evaluation_time = 0.0;
    double detection_evaluation_time = 0.0;

    size_t total_frames = 0;
    
    if ((mkdir("./results/dets", ACCESSPERMS) != 0) && errno != EEXIST){
        cout << "Failed to create output directory." << endl; 
        return -1;
    }
    if ((mkdir("./results/tracks", ACCESSPERMS) != 0) && errno != EEXIST){
        cout << "Failed to create output directory." << endl; 
        return -1;
    }
    // create instance of DPM detector 
    DPM mot_detector;
    // detection result vector format: frames x numDetPerFrame x MOT_Style_row
    //      Note: frames roll over seqeuntially after each seqenuce 
    vector<vector<vector<double>>> detections;
    
    // tracking results vector holds each results of each sequence, combined at end
    vector<trackingResults> track_results(sequences.size());
    for (size_t i = 0; i < sequences.size(); ++i) {
        string& seq = sequences[i];
        // create instance of SORT tracker
        SORT mot_tracker; 

        // collect all image names in sequence
        vector<string> image_names;
        string seq_loc = "data/" + seq + "/img1";
        get_all_files_names_within_folder(seq_loc, image_names);
        std::sort(image_names.begin(), image_names.end());
        
        // prepare vector to collect mot_style trackings for evaultion
        vector<vector<vector<double>>> trackings(image_names.size());
       
        for (int frame = 0; frame < (int)image_names.size(); ++frame) {
            
            // perform detections on this frame
            vector<vector<double>> frame_dets; // data in MOT style, used to calculate detection results
            vector<vector<float>>  SORT_dets; // data in SORT style, used to perform SORT tracking algorithm
            string image_loc = seq_loc + image_names[frame];
            
            clock_t start_time = clock();
            mot_detector.get_detections(image_loc, total_frames, SORT_dets, frame_dets); 
            detection_time += ((float)(clock() - start_time)) / CLOCKS_PER_SEC;

            detections.emplace_back(move(frame_dets));

            // update SORT tracking with new detections
            vector<vector<float> > trackers;
            start_time = clock();
            trackers = mot_tracker.update(SORT_dets);
            track_time += ((float)(clock() - start_time)) / CLOCKS_PER_SEC;
            
            // shift bounding box coords to MOT standard and store in trackins for results
            vector<vector<double> > mot_tracks(trackers.size(), vector<double>(10, -1));
            for(size_t i = 0; i < trackers.size(); ++i) {
                mot_tracks[i][0] = (double)(frame+1); 
                mot_tracks[i][1] = (double)trackers[i][4];
                mot_tracks[i][2] = (double)trackers[i][0];
                mot_tracks[i][3] = (double)trackers[i][1];
                mot_tracks[i][4] = (double)trackers[i][2] - (double)trackers[i][0]; 
                mot_tracks[i][5] = (double)trackers[i][3] - (double)trackers[i][1];
                mot_tracks[i][6] = (double)1.0;            
            }
            trackings[frame] = (move(mot_tracks)); 

            total_frames++;
        }
        // evaluate single seqeunce tracking results. Store in track_results for total computation
        clock_t start_time = clock();
        evalTrack(seq, trackings, track_results[i]);
        track_evaluation_time += ((float)(clock() - start_time)) / CLOCKS_PER_SEC;
        //cout << "done tracking " << seq << endl << endl;
    }
    assert(total_frames == detections.size());
    cout << "\nDone detecting and tracking!" << endl;

    // evaluate all the detection results
    clock_t start_time = clock();
    evalDet(sequences, detections);
    detection_evaluation_time = ((float)(clock() - start_time)) / CLOCKS_PER_SEC;

    // calculate the combined tracking results
    start_time = clock();
    computeTrackResults(track_results);
    track_evaluation_time += ((float)(clock() - start_time)) / CLOCKS_PER_SEC;

    cout << "Detection time: " << detection_time 
         << " for " << total_frames << " frames (" 
         << total_frames / detection_time << " FPS)" << endl;
         
    cout << "Tracking time: " << track_time 
         << " for " << total_frames << " frames (" 
         << total_frames / track_time << " FPS)" << endl;

    cout << "Detection evaluation time: " << detection_evaluation_time 
         << " for " << total_frames << " frames (" 
         << total_frames / detection_evaluation_time << " FPS)" << endl;

    cout << "Tracking evaluation time: " << track_evaluation_time 
         << " for " << total_frames << " frames (" 
         << total_frames / track_evaluation_time << " FPS)" << endl;

    return 0;
}
