#include "sort.hpp"
#include "kalman_box_tracker.hpp"
#include "linear_assignment.hpp"

#include <vector>
#include <utility>
#include <cmath>
using std::vector;
using std::pair;
using std::make_pair;
using std::fmax;
using std::fmin;
using std::isnan;

#include <iostream>
using std::cout;
using std::endl;

float iou(const vector<float> &bb_test,const vector<float> &bb_gt) {
    float xx1 = fmax(bb_test[0], bb_gt[0]);
    float yy1 = fmax(bb_test[1], bb_gt[1]);
    float xx2 = fmin(bb_test[2], bb_gt[2]);
    float yy2 = fmin(bb_test[3], bb_gt[3]);
    float w = fmax(0.0, xx2 - xx1);
    float h = fmax(0.0, yy2 - yy1);
    float wh = w * h;
    float o = wh / ((bb_test[2]-bb_test[0])*(bb_test[3]-bb_test[1])
    + (bb_gt[2]-bb_gt[0])*(bb_gt[3]-bb_gt[1]) - wh);
    return o;
}

void associate_detections_to_trackers(
    /*return values. assumed empty*/
    vector<pair<int,int> > &matches,
    vector<int> &unmatched_detections,
    vector<int> &unmatched_trackers,
    /*inputs*/
    vector<vector<float> > detections,
    vector<vector<float> > trackers,
    float iou_threshold = 0.3
) {
    if (trackers.size() == 0) {
        for (int i = 0; i < (int)detections.size(); ++i) {
            unmatched_detections.push_back(i);
        }
        return;
    }
    vector<vector<float> > iou_matrix(detections.size(), vector<float>(trackers.size(), 0));
    vector<vector<float> > neg_iou_matrix(detections.size(), vector<float>(trackers.size(), 0));
    for (int d = 0; d < (int)detections.size(); ++d){
        for (int t = 0; t < (int)trackers.size(); ++t) {
            iou_matrix[d][t] = iou(detections[d], trackers[t]);
            neg_iou_matrix[d][t] = -1 * iou_matrix[d][t];
        }
    }
    vector<pair<int, int> > matched_indices = linear_assignment(neg_iou_matrix);
    for (int d = 0; d < (int)detections.size(); ++d) {
        bool in_matched = false;
        for (auto m : matched_indices) {
            if (d == m.first) {
                in_matched = true;
                break;
            }
        }
        if (!in_matched) {
            unmatched_detections.push_back(d);
        }
    }
    for (int t = 0; t < (int)trackers.size(); ++t) {
        bool in_matched = false;
        for (auto m : matched_indices) {
            if (t == m.second) {
                in_matched = true;
                break;
            }
        }
        if (!in_matched) {
            unmatched_trackers.push_back(t);
        }
    }
    //filter out matched with low IOU
    for (auto m : matched_indices) {
        if (iou_matrix[m.first][m.second] < iou_threshold) {
            unmatched_detections.push_back(m.first);
            unmatched_trackers.push_back(m.second);
        }
        else {
            matches.push_back(m);
        }
    }
}


SORT::SORT(int _max_age, int _min_hits)
    : max_age(_max_age),
      min_hits(_min_hits),
      trackers(),
      frame_count(0){}

vector<vector<float> > SORT::update(vector<vector<float> > dets) {
    //Params:
    //  dets - a numpy array of detections in the format [[x1,y1,x2,y2,score],[x1,y1,x2,y2,score],...]
    //Requires: this method must be called once for each frame even with empty detections.
    //Returns the a similar array, where the last column is the object ID.
    frame_count++;
    vector<vector<float> > trks(trackers.size(), vector<float>(5, 0));
    vector<int> to_del;
    vector<vector<float> > ret;
    for (int t = 0; t < (int)trks.size(); ++t) {
        vector<float> pos = trackers[t].predict();
        bool has_nan = false;
        for (int i = 0; i < 4; ++i) {
            trks[t][i] = pos[i];
            if (isnan(trks[t][i])) {
               cout << "nan value in tracker: frame " << frame_count
			   		<< " and tracker index " << i << endl << endl; 
			   has_nan = true;
            }
        }
        trks[t][4] = 0;
        if (has_nan) {
            to_del.push_back(t);
        }
    }
    //line 202. we are throwing out any trackers that spit out a nan
    for (int i = to_del.size() - 1; i >= 0; --i) {
        trackers.erase(trackers.begin() + to_del[i]);
        trks.erase(trks.begin() + to_del[i]);
    }
    vector<pair<int,int> > matched;
    vector<int> unmatched_dets;
    vector<int> unmatched_trks;
    associate_detections_to_trackers(matched, unmatched_dets, unmatched_trks,
                                     dets, trks);

    //update matched trackers with assigned detections
    for (int t = 0; t < (int)trks.size(); ++t) {
        bool unmatched = false;
        for (auto i : unmatched_trks) {
            if (i == t) {
                unmatched = true;
                break;
            }
        }
        if (!unmatched) {
           //we know it is matched. grab it's first
           int d = -1;
           for (auto i : matched) {
               if (t == i.second) {
                   d = i.first;
                   break;
               }
           }
           trackers[t].update(dets[d]);
        }
    }

    //create and initialize new trackers for unmatched detections
    for (auto i : unmatched_dets) {
        trackers.push_back(KalmanBoxTracker(dets[i]));
    }
    for (int t = trackers.size() - 1; t >= 0; --t) {
        vector<float> d = trackers[t].get_state();
       // if ((trackers[t].get_time_since_update() < 1) && (trackers[t].get_hit_streak() >= min_hits || frame_count <= min_hits)) {
        if ((trackers[t].get_time_since_update() <= max_age) && (trackers[t].get_hits() >= min_hits || frame_count <= min_hits)) {
            d.push_back((float)trackers[t].get_id() + 1);
            ret.push_back(d);
        }
        //remove dead tracklet
        if (trackers[t].get_time_since_update() > max_age) {
            trackers.erase(trackers.begin() + t);
        }
    }

    return ret;
}
