#include "trackUtils.hpp"

/**
 * Following from MinCostMex
 */

// Min cost bipartite matching via shortest augmenting paths
//
// Code from https://github.com/jaehyunp/
//
// This is an O(n^3) implementation of a shortest augmenting path
// algorithm for finding min cost perfect matchings in dense
// graphs.  In practice, it solves 1000x1000 problems in around 1
// second.
//
//   cost[i][j] = cost for pairing left node i with right node j
//   Lmate[i] = index of right node that left node i pairs with
//   Rmate[j] = index of left node that right node j pairs with
//
// The values in cost[i][j] may be positive or negative.  To perform
// maximization, simply negate the cost[][] matrix.

double MinCostMatching(const VVD &cost, VI &Lmate, VI &Rmate) {
	int n = int(cost.size());

	// construct dual feasible solution
	VD u(n);
	VD v(n);
	for (int i = 0; i < n; i++) {
		u[i] = cost[i][0];
		for (int j = 1; j < n; j++) u[i] = min(u[i], cost[i][j]);
	}
	for (int j = 0; j < n; j++) {
		v[j] = cost[0][j] - u[0];
		for (int i = 1; i < n; i++) v[j] = min(v[j], cost[i][j] - u[i]);
	}

	// construct primal solution satisfying complementary slackness
	Lmate = VI(n, -1);
	Rmate = VI(n, -1);
	int mated = 0;
	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++) {
			if (Rmate[j] != -1) continue;
			if (fabs(cost[i][j] - u[i] - v[j]) < 1e-10) {
				Lmate[i] = j;
				Rmate[j] = i;
				mated++;
				break;
			}
		}
	}

	VD dist(n);
	VI dad(n);
	VI seen(n);

	// repeat until primal solution is feasible
	while (mated < n) {

		// find an unmatched left node
		int s = 0;
		while (Lmate[s] != -1) s++;

		// initialize Dijkstra
		fill(dad.begin(), dad.end(), -1);
		fill(seen.begin(), seen.end(), 0);
		for (int k = 0; k < n; k++)
			dist[k] = cost[s][k] - u[s] - v[k];

		int j = 0;
		while (true) {

			// find closest
			j = -1;
			for (int k = 0; k < n; k++) {
				if (seen[k]) continue;
				if (j == -1 || dist[k] < dist[j]) j = k;
			}
			seen[j] = 1;

			// termination condition
			if (Rmate[j] == -1) break;

			// relax neighbors
			const int i = Rmate[j];
			for (int k = 0; k < n; k++) {
				if (seen[k]) continue;
				const double new_dist = dist[j] + cost[i][k] - u[i] - v[k];
				if (dist[k] > new_dist) {
					dist[k] = new_dist;
					dad[k] = j;
				}
			}
		}

		// update dual variables
		for (int k = 0; k < n; k++) {
			if (k == j || !seen[k]) continue;
			const int i = Rmate[k];
			v[k] += dist[k] - dist[j];
			u[i] -= dist[k] - dist[j];
		}
		u[s] += dist[j];

		// augment along path
		while (dad[j] >= 0) {
			const int d = dad[j];
			Rmate[j] = Rmate[d];
			Lmate[Rmate[j]] = j;
			j = d;
		}
		Rmate[j] = s;
		Lmate[s] = j;

		mated++;
	}

	double value = 0;
	for (int i = 0; i < n; i++)
		value += cost[i][Lmate[i]];

	return value;
}

void MinCostAssignment(VVD& distMatrix, VI& assignment)
{

	int nOfRows    = distMatrix.size();
	int nOfColumns = distMatrix[0].size();

	double total_cost = 0.0;

	if (nOfRows == 0 || nOfColumns == 0)
		return;

	/* Call C-function */
    double INF = 1e8;
    int squareSize = max(nOfRows, nOfColumns);
	// TODO: cut out old "alldist" middleman
	// 		 only question is inf... Handle when adding fp / fn blocks to cost Mat?
	vector<vector<double>> alldist(squareSize, vector<double>(squareSize, INF));
    for (int i = 0; i < nOfRows; i++)
    {
        for (int j =0; j < nOfColumns; j++)
        {
            alldist[i][j] = distMatrix[i][j];
        }
    }
    
	vector<int> Lmate, Rmate;
	MinCostMatching(alldist, Lmate, Rmate);
    
    for (int i = 0; i < nOfRows; i++)
    {
		if (Lmate[i] < nOfColumns && alldist[i][Lmate[i]] != INF)
		{
			total_cost += alldist[i][Lmate[i]];
			// This is a linear compacted mat in MatLab format...
			// assignment[Lmate[i] * nOfRows + i] = 1;
			// Pretty sure this is correct conversion...
			//assignment[i][Lmate[i]] = 1;
	// converted from 2d assigment to 1d
			assignment[i] = Lmate[i];
		}
    }
}


/**
 *  Following from CostBlockMex
 */
void correspondingFrames(VVD& tr1, VVD& tr2, int* loc)
{
	size_t pos = 0;
	size_t i = 0;
	while (i < tr1.size() && pos < tr2.size()) {
		while (pos < tr2.size()) {
			if (tr1[i][0] == tr2[pos][0]) {
				loc[i] = pos; pos++; i++;
				break;
			}
			else if (tr1[i][0] < tr2[pos][0]) {
				loc[i] = -1; i++;
				if (i == tr1.size()) break;
			}
			else pos++;
		}
	}
}

void computeDistances(VVD& tr1, VVD& tr2, int numPoints1, int numPoints2, int* position, double* distance)
{
	for (size_t i = 0; i < tr1.size(); i++)
	{
		if (position[i] == -1)
			distance[i] = 0;
		else
		{
			double area1 = tr1[i][4] * tr1[i][5];
			double area2 = tr2[position[i]][4] * tr2[position[i]][5];

			double x_overlap = max(0.0, min(tr1[i][2] + tr1[i][4], tr2[position[i]][2] + tr2[position[i]][4]) - max(tr1[i][2], tr2[position[i]][2]));
			double y_overlap = max(0.0, min(tr1[i][3] + tr1[i][5], tr2[position[i]][3] + tr2[position[i]][5]) - max(tr1[i][3], tr2[position[i]][3]));
			double intersectionArea = x_overlap*y_overlap;
			double unionArea = area1 + area2 - intersectionArea;
			double iou = intersectionArea / unionArea;
			distance[i] = iou;
		}
	}
}

void computeCostBlock(VVD& tr1, VVD& tr2, double threshold, double& cost, double& fp, double& fn)
{
	size_t numPoints1 = tr1.size();
	size_t numPoints2 = tr2.size();
	int tr1start, tr1end, tr2start, tr2end;
	tr1start = tr1[0][0];
	tr1end = tr1[numPoints1 -1][0];
	tr2start = tr2[0][0];
	tr2end = tr2[numPoints2 -1][0];

	bool overlapTest = ((tr1start >= tr2start && tr1start <= tr2end) ||
						(tr1end >= tr2start && tr1end <= tr2end) ||
						(tr2start >= tr1start && tr2start <= tr1end) ||
						(tr2end >= tr1start && tr2end <= tr1end));
		
	if (!overlapTest) {
		fp = numPoints2;
		fn = numPoints1;
		cost = numPoints1 + numPoints2;
		return;
	}

	int* positionGT = new int[numPoints1];
	int* positionPred = new int[numPoints2];
	for (size_t i = 0; i < numPoints1; i++) positionGT[i] = -1;
	for (size_t i = 0; i < numPoints2; i++) positionPred[i] = -1;
	double* distanceGT = new double[numPoints1];
	double* distancePred = new double[numPoints2];
	
	correspondingFrames(tr1, tr2, positionGT);
	correspondingFrames(tr2, tr1, positionPred);
	computeDistances(tr1, tr2, numPoints1, numPoints2, positionGT, distanceGT);
	computeDistances(tr2, tr1, numPoints2, numPoints1, positionPred, distancePred);

	fp = 0; fn = 0;
	for (size_t i = 0; i < numPoints1; i++) if (distanceGT[i] < threshold) fn++;
	for (size_t i = 0; i < numPoints2; i++) if (distancePred[i] < threshold) fp++;

	cost = fp + fn;

	delete[] positionGT;
	delete[] positionPred;
	delete[] distanceGT;
	delete[] distancePred;
}

// incoming raw GTs and Preds are split into vectors of [object ID x numRows/ID x info_row]
void calcCostBlock(VVVD& gt, VVVD& pred, VVD& cost,	VVD& fp, VVD& fn)
{
	double threshold = 0.5;

	for(size_t g = 0; g < gt.size(); ++g) {
		for (size_t p = 0; p < pred.size(); ++p) {
			// some prediction (i.e. tracking) ids are missing do to preprocess cleaning 
			if(pred[p].size() == 0)
				continue;
			VVD& tr1 = gt[g];
			VVD& tr2 = pred[p];
			computeCostBlock(tr1, tr2, threshold, cost[g][p], fp[g][p], fn[g][p]);
		}
	}
}

/*
 * Following from Clear_MOT_Hun_mex
 */

inline double euclidean(double x1, double y1, double x2, double y2)
{
	return sqrt((x1 - x2)*(x1 - x2) + (y1 - y2)*(y1 - y2));
}

double adjusted_boxiou(double l1, double t1, double w1, double h1, double l2, double t2, double w2, double h2)
{
	double area1 = w1 * h1;
	double area2 = w2 * h2;

	double x_overlap = max(0.0, min(l1 + w1, l2 + w2) - max(l1, l2));
	double y_overlap = max(0.0, min(t1 + h1, t2 + h2) - max(t1, t2));
	double intersectionArea = x_overlap*y_overlap;
	double unionArea = area1 + area2 - intersectionArea;
	double iou = intersectionArea / unionArea;
	return iou;
}

/*
 * The following are used to re-format to data to work better with existing MOT-devkit functions
 */

// converts both gt and pred from [numRows x info_row] to [ids x numRows/ID x info_row]
void formatByID(VVD& gtRaw, VVD& predRaw, VVVD& gt, VVVD& pred){

	for(size_t row = 0; row < gtRaw.size(); ++row) {
		gt[gtRaw[row][1] - 1].emplace_back(gtRaw[row]);
	}

	for(size_t row = 0; row < predRaw.size(); ++row) {
		pred[predRaw[row][1] - 1].emplace_back(predRaw[row]);
	}
}

// converts [frame x numRows/frame x row] to [numRows x row]
// Also removes unused ids (removed by cleaning), and re-labels all
// 	ids to start consecutively from 1 
void formatToLongList(VVVD& rawData, VVD& data) {

	vector<bool> idPresent;
	for(auto& frame : rawData) {
		sort(frame.begin(), frame.end(), ByID());
		for(auto& row : frame) {
			data.emplace_back(row);
			int id = (int)row[1];
			while((int)idPresent.size() < id + 1)
				idPresent.emplace_back(false);
			idPresent[id] = true;
		}
	}
	VI assignment(idPresent.size(), 0);	
	int curId = 1;
	for(int i = 1; i < (int)idPresent.size(); ++i){
		if(idPresent[i]){
			assignment[i] = curId;
			++curId;
		}
	}
	for(VD& row : data)
		row[1] = assignment[(int)row[1]];
	sort(data.begin(), data.end(), ByID());
}
